/*

   // Original
   Pure JavaScript Wiki Like Formatter v0.8
   http://hail2u.net/blog/files/pjswlf.html

*/

var pjswlf = {
	reStrong:  /\'\'\'(.*?)\'\'\'/g,
	reEm:      /\'\'(.*?)\'\'/g,
	reDel:     /\-\-(.*?)\-\-/g,
	reIns:     /\+\+(.*?)\+\+/g,
	reImg:     /\{\{(.*?):(\d+?):(\d+?):(.*?)\}\}/g,

	reExtra:   /\{\{([^\}]*?)\}\}/g,
	reSpChars: /\\([^a-zA-Z0-9])/g,

	doMarkup: function (s, h) {
		if(h) this.ext   = new extraFormat(h);
		var div   = document.createElement("div");
		var lines = s.replace(/\r\n|\r|\n/g, "\n").split(/\n\n/);

		for (var i = 0; i < lines.length; i++) {
			var line = lines[i];
            var elm  = 
				this.markupInline(
					this.markupBlock(line)
				);
			if(elm) div.appendChild(elm);
		}

		return div;
	},

	markupBlock: function (s) {
		var elm, txt;

		if (s.match(/^\n*$/)) {
			return;
		} else if (s.indexOf("****") == 0) {
			elm = document.createElement("h6");
			txt = document.createTextNode(
				s.replace(/^\*\*\*\*\s*/, "")
			);
			elm.appendChild(txt);
		} else if (s.indexOf("***") == 0) {
			elm = document.createElement("h5");
			txt = document.createTextNode(
				s.replace(/^\*\*\*\s*/, "")
			);
			elm.appendChild(txt);
		} else if (s.indexOf("**") == 0) {
			elm = document.createElement("h4");
			txt = document.createTextNode(
				s.replace(/^\*\*\s*/, "")
			);
			elm.appendChild(txt);
		} else if (s.indexOf("*") == 0) {
			elm = document.createElement("h3");
			txt = document.createTextNode(
				s.replace(/^\*\s*/, "")
			);
			elm.appendChild(txt);
		} else if (s.indexOf("----") == 0) {
			elm = document.createElement("hr");
		} else if (s.indexOf(">") == 0) {
			elm = document.createElement("blockquote");

			//var aQuotes = s.replace(/^>\s*/gm, "").split(/\n/);
			var aQuotes = s.split(/^>\s*/gm);

			for (var j = 0; j < aQuotes.length; j++) {
				elm.appendChild(this.doMarkup(aQuotes[j]));
			}
		} else if (s.indexOf(" ") == 0) {
			elm = document.createElement("div");

			var aPres = s.replace(/^\s/gm, "").split(/\n/);
			var pre = document.createElement("pre");
			for (var j = 0; j < aPres.length; j++) {
				pre.appendChild(document.createTextNode(aPres[j]));
				pre.appendChild(document.createTextNode('\n'));
			}
			elm.appendChild(pre);
		} else if (s.indexOf("-") == 0) {
			elm = this.markupListed(s, 'ul', /^\-/);
		} else if (s.indexOf("+") == 0) {
			elm = this.markupListed(s, 'ol', /^\+/);

		} else if (s.indexOf(":") == 0) {
			elm = document.createElement("dl");
			var aLi = s.replace(/^\:\s*/gm, "").split(/\n/);

			for (var j = 0; j < aLi.length; j++) {
				if (aLi[j].indexOf(":") != 0) {
					var dt = document.createElement("dt");
					txt = document.createTextNode(aLi[j]);
					dt.appendChild(txt);
					elm.appendChild(dt);
				} else {
					var dd = document.createElement("dd");
					txt = document.createTextNode(aLi[j].replace(/^\:\s*/, ""));
					dd.appendChild(txt);
					elm.appendChild(dd);
				}
			}
		} else {
			elm = document.createElement("p");
			txt = document.createTextNode(s);
			elm.appendChild(txt);
		}

		return elm;
	},

	markupInline: function (elm) {
		if(!elm) return;
		if (elm.childNodes.length > 1) {
			for (var i = 0; i < elm.childNodes.length; i++) {
				this.markupInline(elm.childNodes[i]);
			}
		} else {
			elm = this._txt2html(elm);
		}

		return elm;
	},

	_txt2html: function (elm) {
		if(!elm) return;
		var s = elm.innerHTML;

		if (s != "") {
			s = s.replace(this.reStrong,  '<strong>$1</strong>');
			s = s.replace(this.reEm,      '<em>$1</em>');
			s = s.replace(this.reDel,     '<del>$1</del>');
			s = s.replace(this.reIns,     '<ins>$1</ins>');
			//s = s.replace(this.reImg,     '<img src="$4" alt="$1" width="$2" height="$3" />');
			s = this.ext.parseLink(s);
			s = s.replace(this.reSpChars, '$1');
			elm.innerHTML = s;
		}

		return elm;
	},

   markupListed: function(txt, tag, key) {
      txt = txt.replace(/\n$/, '');
      this.s   = txt.split(/\n/);
      this.j   = 0;
      this.lvl = 1;

      this.nested = function() {
         var ul = document.createElement(tag);

         while(this.j < this.s.length) {
            var t   = this.s[this.j];
            var lvl = 0;

            while(t.match(key)) {
               t = t.replace(key, '');
               lvl++
            }
            if(lvl==0) lvl=1;

            if(lvl>this.lvl) {
               this.lvl++;
               ul.appendChild(this.nested());
            } else if(lvl<this.lvl) {
               this.lvl--;
               break;
            } else {
               var li = document.createElement("li");
               li.appendChild(document.createTextNode(t));
               ul.appendChild(li);
               this.j++;
            }
         }
         return ul;
      }

      return this.nested();
   }
};
