/*
 *  ScrapMemo - 候補補完ユーティリティ
 *
 *  Copyright (c) 2005, Kyosuke Takayama <support at mc.neweb.ne.jp> 
 *  http://espion.just-size.jp/
 *
 *  $Id: suggest.js 11 2005-10-07 01:38:07Z takayama $
*/

/*
   * フォームを作成
   new SCM.suggest(
         document.sForm.search,     // inputフィールド
         makeList,                  // 補完候補を作成する関数
         doSearch,                  // submit 時の動作を設定
         'suggest'                  // クラス名 (無指定時は suggest)
   );

   * 複数のフォームで使いたかったら
   new SCM.suggest を複数回やるだけ

*/

if(typeof(SCM) == 'undefined') var SCM = {}

// フォームセッティング
function _SCM_suggest(input, list, submit, cls) {
   this._timer = '';
   this.nlist  =  new Array();
   this.slist  =  new Array();

   // 外部用
   this.submit   = function() { return true;  };
   this.makelist = function() { return false; };

   var _this = this;

   if(submit) this.submit   = submit;
   if(list)   this.makelist = list;

   this.suggestInput = input;
   this.suggestElm   = this.createForm(cls);

   input.form.onsubmit = this.submit;
   this.suggestInput.setAttribute('autocomplete', 'off');
   this._addEvent(this.suggestInput, 'focus',    function()  { _this.init(); });
   this._addEvent(this.suggestInput, 'blur',     function()  { _this.stop(); });
   this._addEvent(this.suggestInput, 'keypress', function(e) { _this.keypress(e); });
   this._addEvent(this.suggestInput, 'keyup',    function(e) { _this.keyup(e);    });
}
SCM.suggest = _SCM_suggest;

// init - データ初期化
SCM.suggest.prototype.init = function() {
   this.oldVal = '';
   this.oldNum = -1;
   this.maxNum = -1;

   //this.slist = this.makelist();

   this._start();
}

SCM.suggest.prototype.createForm = function(cls) {
   var input = this.suggestInput;
   var fname = input.form;
   var elm = document.createElement('div');
   elm.className = cls||'suggest';
   elm.style.border = '1px solid #999';
   elm.style.backgroundColor = '#FFFFFF';
   elm.style.position = 'absolute';
   elm.style.padding = '0px 3px 0px 3px';
   elm.style.zIndex  = 1;
   elm.style.display = 'none';
   elm.style.width = (document.all && !window.opera) ? input.offsetWidth+'px' : (input.offsetWidth-8)+'px';
   elm.style.top  = input.offsetTop + input.offsetHeight + 'px';
   elm.style.left = '0px';

   elm.appendChild(document.createTextNode(''));
   fname.appendChild(elm);

   return elm;
}

// _start - 補完開始
SCM.suggest.prototype._start = function() {
   var _this = this;
   if(!this._timer) this._timer  = window.setInterval(function(){_this.doSuggest();}, 400);
}

// keypress - エンターキー取得
SCM.suggest.prototype.keypress = function(e) {
   if(e) {
      if(e.keyCode == 13 && this.oldNum >= 0)
      this.suggestInput.value = this.nlist[this.oldNum];
   }
   this._start();
}

// keyup - カーソルキー専用処理
// (IE だと keypress で カーソルキーの keycode が取得できないようなので)
SCM.suggest.prototype.keyup = function(e) {
   if(e) {
      var code = e.keyCode;
      if(code == 38) this._up();
      else if(code == 40) this._down();
   }
   this._start();
}

// stop - 処理終了
SCM.suggest.prototype.stop = function() {
   if(this._timer) clearTimeout(this._timer);
   this._timer  = '';
   this.oldVal = '';
   this._hide();
}

// mover - マウスオーバー
SCM.suggest.prototype.mover = function(num) {
   if(this.oldNum >= 0) document.getElementById('sug'+this.oldNum).className = 'off';
   document.getElementById('sug'+num).className = 'on';
   this.oldNum = num;
}

// setword - 候補確定
SCM.suggest.prototype.setword = function(word) {
   this.suggestInput.value = word;
   this.submit();
}

// doEvent - イベント処理
SCM.suggest.prototype.doEvent = function(_this, event, param) {
   return function() { _this[event](param); }
}

// doSuggest
SCM.suggest.prototype.doSuggest = function() {
   var val = this.suggestInput.value;
   if(val != this.oldVal && val) {
      if(!this.oldVal) this._show();
      this.oldVal = val;
      this.oldNum = -1;
      this.maxNum = -1;

      var ol = document.createElement('ol');
      this.nlist = this.makelist(val);

      for(var i in this.nlist) {
         var word = this.nlist[i];
         var li = document.createElement('li');
         li.className = 'off';
         li.setAttribute('id', 'sug'+i);

         var a  = document.createElement('a');
         a.appendChild(document.createTextNode(word));
         a.setAttribute('href', 'javascript:;');
         this._addEvent(a, 'mouseover', new this.doEvent(this, 'mover', i));
         this._addEvent(a, 'mousedown', new this.doEvent(this, 'setword', word));

         li.appendChild(a);
         ol.appendChild(li);

         if(i == 10) break;
      }
      this.maxNum = i;

      if(this.nlist.length == 0) this.stop();
      else if(i == 0 && this.nlist[0] == val) { this.maxNum = -1; this.stop(); }
      else this.suggestElm.replaceChild(ol, this.suggestElm.firstChild);
   }
}

// _addEvent - イベント追加
SCM.suggest.prototype._addEvent = function(elm, type, event) {
   if(elm.addEventListener) 
      elm.addEventListener(type, event, false);
   else
      elm.attachEvent('on'+type, event);
}

// _up - キーアップ
SCM.suggest.prototype._up = function() {
   if(this.maxNum < 0) return;
   var num = this.oldNum - 1;
   if(0 > num) num = this.maxNum;
   this.mover(num);
}

// _down - キーダウン
SCM.suggest.prototype._down = function() {
   if(this.maxNum < 0) return;
   var num = this.oldNum + 1;
   if(this.maxNum < num) num = 0;
   this.mover(num);
}

// _show - 表示
SCM.suggest.prototype._show = function() {
   this.suggestElm.style.display = 'block';
}

// _hide - 隠蔽
SCM.suggest.prototype._hide = function() {
   this.suggestElm.style.display = 'none';
}

