#! /usr/bin/env perl
#! /usr/bin/speedy -- -t90000

# ScrapMemo
#
# Copyright (c) 2005, Kyosuke Takayama <support at mc.neweb.ne.jp>
# http://espion.just-size.jp/
#
# $Id: smemo.cgi 4 2005-09-20 05:25:07Z takayama $

#use CGI::Carp qw(fatalsToBrowser);

my $scm = SCM->new(
   max   => 30,    # 一覧
   max2  => 50,    # 検索結果
   smax  => 50,    # 検索履歴
   bmax  => 20,    # バックアップ件数

   # RSS 出力用
   URL   => 'http://espion.just-size.jp/files/js/smemo/',
   TITLE => 'ScrapMemo',
   DESCRIPTION => 'ScrapMemo',
);
$scm->run;

# SCM
package SCM;

use strict;

sub new {
   my $pkg  = shift;
   my %opt  = @_;

   my $self = { 
      util  => new SCM::Util,
      max   => 30,
      max2  => 50,
      smax  => 50,
      bmax  => 20,
      TITLE => 'ScrapMemo',
   };

   foreach (keys %opt) {
      $self->{$_} = $opt{$_};
   }

   bless $self, $pkg;
}

sub run {
   my $self = shift;

   print $self->{util}->header(
      -type   => 'text/xml',
      -status => '200 OK',
      -charset=> 'UTF-8',
   );

   $self->{util}->lock;

   if($self->{util}->getParam('name')) {
      $self->modFile();
   } elsif($self->{util}->getParam('dq')) {
      $self->delSearch();
   } else {
      $self->doSearch();
   }

   $self->{util}->unlock;
}

sub doSearch {
   my $self = shift;

   my $result = $self->searchFile($self->{util}->getParam('q'));
   my $xml    = SCM::XML->new;
   my $tree   = $xml->init('smemoindex');

   splice(@{$result->{lists}}, $self->{max2});

   $tree->{smemo} = $result->{lists};
   $tree->{tags}  = $result->{tags};

   $xml->pout;
}

sub searchFile {
   my @lists;
   my $grep;
   my $self = shift;
   my $word = shift;

   #$word =~ tr/!"#$%&'=~//d;
   $word =~ tr/$&'"()|\\[]{}//d;
   $word = 'T' if(!$word or length($word) < 2);
   $grep = $1 if($word =~ /^g (.*)/);

   my @words;
   my @tWords;
   foreach my $w (split(/\s/, $word)) {
      if($w =~ /^tag:(.*)/) {
         push(@tWords, $1);
         $w = "<tag>$1</tag>";
      }
      push(@words, $w);
   }

   my %tags;
   my $xml  = SCM::XML->new;
   opendir(DIR, "memo");
   while (my $file = readdir(DIR)) {
      next if($file !~ /^\d+.xml$/);
      my $value;
      my $flag;
      my $i;
      if($grep) {
         my $name = $file;
         $name =~ s/.xml$//;
         open(IN, "memo/$file");
         while(<IN>) {
            if(/$grep/i) {
               my $push;
               $push->{name}[0]    = $name;
               $push->{subject}[0] = $_;
               $push->{date}[0]    = $self->{util}->getTime((stat("memo/$file"))[9]);
               push(@lists, $push)
            }
         }
         close(IN);
      } else {
         open(IN, "memo/$file");
         while(<IN>) {
            $value .= $_;
         }
         close(IN);

         map { $flag ++ if($value =~ /$_/i); } @words;

         if($flag and $flag == $#words+1) {
            my $tree = $xml->read($value);
            my $push;
            $push->{name}[0]    = $tree->{name}[0];
            $push->{subject}[0] = $tree->{subject}[0]||'';
            $push->{date}[0]    = $tree->{date}[0];
            map { $tags{lc($_)}++; } @{$tree->{tags}[0]->{tag}};
            push(@lists, $push)
         }
      }
   }
   closedir(DIR);

   my @idx = (0..$#lists);
   @idx = sort { $lists[$b]->{'date'}[0] cmp $lists[$a]->{'date'}[0] || $a <=> $b } @idx;
   @lists = @lists[@idx];

   $self->modSearch($word) if(@lists and $word ne 'T');

   my (@tags, $i);
   foreach my $t (sort { $tags{$b} <=> $tags{$a} } keys %tags) {
      my $push;
      next if(grep($_ eq $t, @tWords));
      #$push->{tag}[0]   = $t;
      push(@tags, $t);
      last if($i++ and $i >= 20);
   }

   _modTag(\@tWords) if(@lists and @tWords);

   return { lists => \@lists, tags => [{ tag => \@tags }] };
}

sub modRSS {
   my $self = shift;
   my @list = @_;
   my $wt;
   my $wtb;
   splice(@list, 15);

   $wt .= <<_XML_;
<?xml version="1.0" encoding="UTF-8" ?>
<rdf:RDF
 xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#"
 xmlns="http://purl.org/rss/1.0/"
 xmlns:dc="http://purl.org/dc/elements/1.1/"
 xmlns:content="http://purl.org/rss/1.0/modules/content/"
 xml:lang="ja"
>
<channel rdf:about="$self->{URL}">

   <title>$self->{TITLE}</title>
   <link>$self->{URL}</link>
   <description>$self->{DESCRIPTION}</description>
   <items>
      <rdf:Seq>
_XML_
   foreach my $xml (@list) {
      my $subj = $xml->{subject}[0]||'untitled';
      $wt  .= "         <rdf:li rdf:resource=\"$self->{URL}#$xml->{name}[0]\" />\n";
      $wtb .= <<_XML_;
<item rdf:about="$self->{URL}#$xml->{name}[0]">
   <title>$subj</title>
   <link>$self->{URL}#$xml->{name}[0]</link>
   <dc:date>$xml->{date}[0]</dc:date>
</item>
_XML_
   }

   $wt .= <<_XML_;
      </rdf:Seq>
   </items>
</channel>
$wtb
</rdf:RDF>
_XML_

   my $file   = "memo/rss.rdf";

   open(OT, ">${file}.bak");
   print OT $wt;
   close(OT);
   rename("${file}.bak", $file);
}

sub modFile {
   my $self = shift;
   my $name = $self->{util}->getParam('name');
   my $body = $self->{util}->getParam('body');
   $body =~ s/(.*?)\n//;
   my $subject = ($1||'');

   $name =~ tr/0-9//dc;
   return if($name eq '');

   my $file   = "memo/${name}.xml";
   my $bfile  = "backup/${name}";
   my $mod    = $self->{util}->getTime(time());

   my $xml    = SCM::XML->new($file);

   my $tree = (-e $file) ? $xml->read : $xml->init('smemo');
   my @tags = split(/\s+/, $self->{util}->getParam('tags'));
   my @nTags;
   my @dTags;
   my $backup;
   if(-e $file and 
      ((!-e $bfile.'.bak1.xml') or (time() - (stat($bfile.'.bak1.xml'))[9] > 3600)) ) {
      $backup = $self->{util}->getTime(time());
   }

   if($self->{util}->getParam('cmd') eq 'add') {
      $subject = $tree->{subject}[0];
      $body    = $tree->{body}[0] . "\n" . $self->{util}->getParam('add');
      @tags    = @{$tree->{tags}[0]->{tag}} if($tree->{tags}[0]);
   }

   $tree->{create}[0]  = $mod if(!$tree->{create}[0]);
   $tree->{date}[0]    = $mod;
   $tree->{name}[0]    = $name;
   $tree->{subject}[0] = $subject;
   $tree->{body}[0]    = $body;

   my %ct;
   @tags = grep(!$ct{lc($_)}++, @tags);

   foreach my $tag (@tags) {
      my $exist;
      for(my $i = 0;$i <= $#{$tree->{tags}[0]->{tag}}; $i++) {
         if($tag eq ${$tree->{tags}[0]->{tag}}[$i]) {
            $tree->{tags}[0]->{tag}[$i] = '';
            $exist = 1;
            last;
         }
      }
      push(@nTags, $tag) if(!$exist);
   }

   my @backups;
   if($backup) {
      my $i;
      push(@backups, $backup);
      $i++;

      foreach (@{$tree->{backups}[0]->{backup}}) {
         push(@backups, $_);
         last if($i++ and $i >= $self->{bmax});
      }

      my $f;
      for($f = ($i-1);0 < $f;$f--) {
         rename("${bfile}.bak${f}.xml", "${bfile}.bak".($f+1).".xml");
      }
      utime(time(), time(), $file);
      rename($file, "${bfile}.bak1.xml");
      @{$tree->{backups}[0]->{backup}} = @backups;
   }

   map { push(@dTags, $_) if($_) } @{$tree->{tags}[0]->{tag}};

   @{$tree->{tags}[0]->{tag}} = @tags;

   $xml->pout();
   $xml->write();

   $self->modList($tree);

   _modTag(\@nTags, 1) if(@nTags);
   _modTag(\@dTags, 2) if(@dTags);
}

## taglist.xml 更新
sub _modTag {
   my $tags = shift;
   my $mode = shift;
   my $file = "memo/taglist.xml";

   my $xml  = SCM::XML->new($file);
   my $tree = $xml->read;
   my @after;
   my @begin;
   my @lists;

   my $i = 1;
   foreach my $rel (@{$tree->{tags}}) {
      my $hit;
      my $tag   = $rel->{tag}[0];
      my $count = $rel->{count}[0];

      for(my $f = 0;$f <= $#{$tags};$f++) {
         if(lc($tag) eq lc($tags->[$f])) {
            $hit = 1;
            $tags->[$f] = '';
            if($mode) { $rel->{count}[0] += ($mode == 1) ? 1 : -1; }
            push(@lists, $rel) if($rel->{count}[0] > 0);
            last;
         }
      }

      push(@after, $rel) if(!$hit);
   }

   # 新規の場合
   if($mode == 1) {
      foreach (@{$tags}) {
         next if(!$_);
         my %data;
         $data{tag}[0]   = $_;
         $data{count}[0] = 1;
         push(@begin, \%data);
      }
   }

   push(@lists, @after);
   push(@begin, @lists);
   @{$tree->{tags}} = @begin;

   $xml->write();
}

## list.xml 更新
sub modList {
   my $self = shift;
   my $new  = shift;
   my $file = "memo/list.xml";

   my $xml  = SCM::XML->new($file);
   my $tree = $xml->read;
   my @lists;

   my %mod;
   map { $mod{$_}[0] = $new->{$_}[0]; } qw(date subject name);

   push(@lists, \%mod);
   map { push(@lists, $_) if($new->{name}[0] != $_->{name}[0]); } @{$tree->{smemo}};
   splice(@lists, $self->{max}) if($self->{max} > 0);

   @{$tree->{smemo}} = @lists;

   $xml->write();

   $self->modRSS(@lists);
}

## search.xml 更新
sub modSearch {
   my $self = shift;
   my $word = shift;
   my $del  = shift;
   my $file = "memo/search.xml";

   return if($word =~ /^tag:\S+$/);

   my $xml  = SCM::XML->new($file);
   my $tree = $xml->read;
   my @lists;

   push(@lists, $word) unless($del);
   map { push(@lists, $_) if(lc($word) ne lc($_)); } @{$tree->{item}};
   splice(@lists, $self->{smax});

   $lists[0] = 'ScrapMemo' if($#lists == -1);
   @{$tree->{item}} = @lists;

   $xml->write();
   $xml->pout() if($del);

   $self->{max} = $self->{max2};
}

sub delSearch {
   my $self = shift;
   my $word = $self->{util}->getParam('dq');

   $self->modSearch($word, 1);
}

# SCM::XML
package SCM::XML;

use strict;
use XML::LibXML;

sub new {
   my $pkg  = shift;
   my $self = {
      file    => shift,
      parser  => XML::LibXML->new,
      docTree => XML::LibXML::Document->new('1.0', 'UTF-8'),
   };

   bless $self, $pkg;
}

sub read {
   my $self = shift;
   my $val  = shift;

   undef $self->{root};
   undef $self->{tree};

   my $dom  = ($val) ? $self->{parser}->parse_string($val) : $self->{parser}->parse_file($self->{file});
   my $node = $dom->documentElement;

   $self->{root} = $node->nodeName;
   $self->{tree} = &_elmToTree($node);

   return $self->{tree};
}

sub init {
   my $self = shift;
   $self->{root} = shift;

   # TODO こんな変な処理は嫌だなぁ
   my %data;
   $self->{tree} = \%data;

   return $self->{tree};
}

sub make {
   my $self = shift;
   return if(defined($self->{docCreated}));

   my $elm = $self->{docTree}->createElement($self->{root});
   $self->{docTree}->setDocumentElement($elm);

   &_treeToElm($self->{docTree}, $elm, $self->{tree});

   $self->{docCreated} = 1;
}

sub write {
   my $self = shift;
   return if(!$self->{file});
   #rename($self->{file}, $self->{file}.'.backup');
   $self->make;
   $self->{docTree}->toFile($self->{file}, 1);
}

sub pout {
   my $self = shift;
   binmode(STDOUT, ":utf8");
   print $self->string;
}

sub string {
   my $self = shift;
   $self->make;
   return $self->{docTree}->toString(1);
}

sub _treeToElm {
   my $doc  = shift;
   my $dom  = shift;
   my $tree = shift;

   foreach my $key (keys %{$tree}) {
      foreach my $data (@{$tree->{$key}}) {
         if(defined(%{$data})) {
            my $elm  = $doc->createElement($key);
            &_treeToElm($doc, $elm, $data);
            $dom->appendChild($elm);
         } else {
            $dom->appendTextChild($key, $data);
         }
      }
   }
}

sub _elmToTree {
   my %dist;
   my $node   = shift;
   return unless( $node->nodeType eq &XML_ELEMENT_NODE );

   foreach my $child ($node->getChildnodes) {
      my $data;

      if($child->hasChildNodes) {
         $data = &_elmToTree($child);

         push(@{$dist{$child->nodeName}}, $data);
      } else {
         if($child->nodeType eq &XML_TEXT_NODE) {
            if($child->nodeValue !~ /^[\s\n]+$/) {
               $data = $child->textContent;
               utf8::encode($data);
               return $data;
            }
         }
      }
   }

   return \%dist;
}

# SCM::Util
package SCM::Util;

use CGI;
#use Jcode;
use JEncode;
use strict;

sub new {
   my $pkg  = shift;
   my $self = {
      cgi   => new CGI
   };

   bless $self, $pkg;
}

sub getTime {
   my $self = shift;
   my $time = shift;
   my ($sec,$min,$hour,$day,$mon,$year,$wday) = localtime($time);
   $year += 1900;
   $mon++;
   my $date = sprintf("%04d-%02d-%02dT%02d:%02d:%02d+09:00", $year,$mon,$day,$hour,$min,$sec);

   return $date;
}

sub sanitize {
   my $self = shift;
   my $str = shift;
   $str =~ s/\x0D//g;
   return $str;

   $str =~ s/&/&amp;/g;
   $str =~ s/</&lt;/g;
   $str =~ s/>/&gt;/g;
   $str =~ s/"/&quot;/g;
   $str =~ s/'/&apos;/g;
   $str =~ s/,/&#44;/g;
   return $str;
}

sub getParam {
   my $self = shift;
   my $param = shift;

   # Jcode よりも JEncode の方が精度が良い様子
   $self->{param}->{$param} =
   JEncode->new(
      $self->sanitize(
         $self->{cgi}->param($param)
      )
   )->utf8
   if(!$self->{param}->{$param});

   return $self->{param}->{$param};
}

sub header {
   my $self = shift;

   return $self->{cgi}->header(@_);
}

sub lock {
   open(UtilLock, ">memo/lock");
   flock(UtilLock, 2);
}

sub unlock {
   close(UtilLock);
}

